/* ascii-casemap.cc - ascii-casemap comparator implementation
 * Rob Earhart
 * $Id: ascii-casemap.cc,v 1.2 1997/12/14 23:17:22 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <ctype.h>
#include "compare.h"
#include "iterator.h"
#include "init.h"

class StringCaseIterator
  : public random_access_iterator<String::value_type, String::difference_type> {
public:
    StringCaseIterator() {}
    StringCaseIterator(const String::const_iterator &_it) : it(_it) {}
    StringCaseIterator(const StringCaseIterator &csi) : it(csi.it) {}
    StringCaseIterator &operator=(const String::const_iterator &_it) {
      it = _it;
      return *this;
    }
    StringCaseIterator &operator=(const StringCaseIterator &csi) {
      it = csi.it;
      return *this;
    }
    StringCaseIterator &operator+=(difference_type d) {
      it += d;
      return *this;
    }
    StringCaseIterator operator+(difference_type d) const {
      StringCaseIterator tmp = it;
      tmp += d;
      return tmp;
    }
    StringCaseIterator &operator-=(difference_type d) {
      it -= d;
      return *this;
    }
    StringCaseIterator operator-(difference_type d) const {
      StringCaseIterator tmp = it;
      tmp -= d;
      return tmp;
    }
    difference_type operator-(const StringCaseIterator &sci) const {
      return it - sci.it;
    }
    StringCaseIterator& operator++() {
      ++it;
      return *this;
    }
    StringCaseIterator operator++(int) {
      StringCaseIterator tmp = it;
      ++it;
      return tmp;
    }
    StringCaseIterator& operator--() {
      --it;
      return *this;
    }
    StringCaseIterator operator--(int) {
      StringCaseIterator tmp = it;
      --it;
      return tmp;
    }
    value_type operator[](difference_type d) const {
      return tolower(it[d]);
    }
    value_type operator*() const {
      return tolower(*it);
    }
    bool operator<(const StringCaseIterator &sci) const {
      return it < sci.it;
    }
    bool operator==(const StringCaseIterator &sci) const {
      return it == sci.it;
    }
    bool operator!=(const StringCaseIterator &sci) const {
      return it != sci.it;
    }
private:
    String::const_iterator it;
};

// The following two functions are necessary to make g++ happy
inline String::difference_type* 
distance_type(const StringCaseIterator&) {
  return (String::difference_type*)(0);
}

inline random_access_iterator_tag
iterator_category(const StringCaseIterator&) {
    return random_access_iterator_tag();
}

// ... and finally, we can go ahead with our comparator functions...
static int OrdAsciiCasemap(const String &first, const String &second) {
  return lexicographical_compare_3way(StringCaseIterator(first.begin()),
				      StringCaseIterator(first.end()),
				      StringCaseIterator(second.begin()),
				      StringCaseIterator(second.end()));
}

static int OrdAsciiCasemapReverse(const String &first, const String &second) {
  return -OrdAsciiCasemap(first, second);
}

static bool EqAsciiCasemap(const String &a, const String &b) {
  return equal(StringCaseIterator(a.begin()),
	       StringCaseIterator(a.end()),
	       StringCaseIterator(b.begin()));
}

static bool MatchAsciiCasemap(const String &a, const String &b, bool anchor) {
  // Someday, this ought to do something intelligent...
  if (anchor) {
    if (a.end() - a.begin() < b.end() - b.begin())
      return false;
    return equal(StringCaseIterator(a.begin()),
		 StringCaseIterator(a.end()),
		 StringCaseIterator(b.begin()));
  } else
    return search(StringCaseIterator(a.begin()),
		  StringCaseIterator(a.end()),
		  StringCaseIterator(b.begin()),
		  StringCaseIterator(b.end())) != StringCaseIterator(a.end());
}

static void SetupAsciiCasemap() {
  Compare::Family *family = Compare::Instance()->FindFamily("i");
  family->Orderings()["i;ascii-casemap"] = &OrdAsciiCasemap;
  family->Orderings()["+i;ascii-casemap"] = &OrdAsciiCasemap;
  family->Orderings()["-i;ascii-casemap"] = &OrdAsciiCasemapReverse;
  family->Equalities()["i;ascii-casemap"] = &EqAsciiCasemap;
  family->Matches()["i;ascii-casemap"] = &MatchAsciiCasemap;
  family->Comparators().insert("i;ascii-casemap");
#ifdef EUDORA_HACKS
  // #*&(! Eudora uses en-nocase...
  family->Orderings()["en-nocase"] = &OrdAsciiCasemap;
  family->Orderings()["+en-nocase"] = &OrdAsciiCasemap;
  family->Orderings()["-en-nocase"] = &OrdAsciiCasemapReverse;
  family->Equalities()["en-nocase"] = &EqAsciiCasemap;
  family->Matches()["en-nocase"] = &MatchAsciiCasemap;
  family->Comparators().insert("en-nocase");
#endif // EUDORA_HACKS
}

static Init init(&SetupAsciiCasemap);
