/* compare.h -*- C++ -*- comparator interface
 * Rob Earhart
 * $Id: compare.h,v 1.4 1998/02/18 21:16:54 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_COMPARE_H
#define ACAP_COMPARE_H

#include <pthread.h>
#include <hash_map.h>
#include <hash_set.h>
#include "str.h"
#include "mutex.h"
#include "grab.h"

// Compare is a singleton class which contains all the information
// about the various comparators available.  Each Family object
// represents a family of comparators (it's broken up this way to make
// it easy for each language to look up the set of comparators
// appropriate to the language).

class Compare {
public:
  static Compare *Instance();

  typedef int (*ordering_t)(const String &first, const String &second);
  typedef bool (*equality_t)(const String &a, const String &b);
  typedef bool (*match_t)(const String &string, const String &pattern, bool anchor);
  typedef hash_map<String, ordering_t,
                   hash<String>, equal_to<String> > ordmap_t;
  typedef hash_map<String, equality_t,
                   hash<String>, equal_to<String> > eqmap_t;
  typedef hash_map<String, match_t,
                   hash<String>, equal_to<String> > matchmap_t;
  typedef hash_set<String, hash<String>, equal_to<String> > comparatorset_t;
  
  class Family {
  public:
    ordmap_t &Orderings() { return orderings; }
    eqmap_t &Equalities() { return equalities; }
    matchmap_t &Matches() { return matches; }
    comparatorset_t &Comparators() { return comparators; }
  private:
    ordmap_t orderings;
    eqmap_t equalities;
    matchmap_t matches;
    comparatorset_t comparators;
  };
  typedef hash_map<String, Family *,
                   hash<String>, equal_to<String> > familymap_t;

  Family *FindFamily(const String &family) {
    Grab g(family_lock);
    pair<familymap_t::iterator, bool> p
      = families.insert(familymap_t::value_type(family, (Family *)NULL));
    if (p.second)
      (*p.first).second = new Family();
    return (*p.first).second;
  }

  ordering_t LookupOrdering(const String &name);
  equality_t LookupEquality(const String &name);
  match_t LookupMatch(const String &name);

private:
  Family *FindFamilyOfComparator(const String &name);
  familymap_t families;
  Mutex family_lock;

  static void Initialize();
  Compare() {}
  static Compare *instance;
  static pthread_once_t instance_init;
};

#endif // ACAP_COMPARE_H
