/* conn.cc - connection implementation
 * Rob Earhart
 * $Id: conn.cc,v 1.10 1998/03/09 00:57:59 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include "auth.h"
#include "conn.h"
#include "ptr.h"
#include "str.h"
#include "acap.h"
#include "dispatch.h"
#include "debug.h"
#include "compare.h"
#include "deletedsince.h"
#include "store.h"
#include "search.h"
#include "error.h"
#include "language.h"
#include "updatecontext.h"
#include "response.h"
#include "notifycontext.h"
#include "re.h"

// This is an rfc1808-compliant URL recognition DFA
static RegExp url_dfa("^([[:alnum:]+.-]+:([-[:alnum:]$_.+!*'(),;/?:@&=]|%[[:xdigit:]]{2})|([[:alnum:]+.-]+:)?((//([-[:alnum:]$_.+!*'(),:@&=;?]|%[[:xdigit:]]{2})*)?/)?(([-[:alnum:]$_.+!*'(),:@&=]|%[[:xdigit:]]{2})+([-[:alnum:]$_.+!*'(),:@&=/]|%[[:xdigit:]]{2})*)?(;([-[:alnum:]$_.+!*'(),:@&=/]|%[[:xdigit:]]{2})*)*(\\?([-[:alnum:]$_.+!*'(),]|%[[:xdigit:]]{2})*)?)(#([-[:alnum:]$_.+!*'(),;/?:@&=]|%[[:xdigit:]]{2})*)?$");

Conn::Conn(int _fd, struct sockaddr_in _peer)
  : fd(_fd), peer(_peer), in(fd), out(fd), lexer(this, &in, &out),
    putback(false), authentic(false), language(Language::Default()) {
      out << "* ACAP (ContextLimit \"100\")"
	     " (Implementation \"Carnegie Mellon Cyrus ACAPd " VERSION "\")"
	     " (SASL ";
      bool dospace = false;
      for (Auth::MechMap::iterator mit = Auth::Mechs().begin();
	   mit != Auth::Mechs().end();
	   ++mit) {
	if (dospace)
	  out << ' ';
	else
	  dospace = true;
	out << '\"' << (*mit).first << '\"';
      }
      out << ")\r\n" << flush;
}

Conn::~Conn() {
  out.flush();
  close(fd);
  DEBUG("Connection closed.");
}

int Conn::NextToken() {
  if (putback)
    putback = false;
  else
    lasttoken = lexer.yylex();
  return lasttoken;
}

void Conn::Resync() {
  if (lasttoken != Token::CRLF && lasttoken != Token::GARBAGE) {
    lexer.Resync();
  }
}

static void eat_space(Conn *conn) {
  if (conn->NextToken() != Token::SPACE)
    throw Response::BAD(&Msg::ExpectedSpace);
}

static void eat_crlf(Conn *conn) {
  if (conn->NextToken() != Token::CRLF)
    throw Response::BAD(&Msg::ExpectedCRLF);
}

static void eat_lparen(Conn *conn) {
  if (conn->NextToken() != Token::LPAREN)
    throw Response::BAD(&Msg::ExpectedLParen);
}

static void eat_rparen(Conn *conn) {
  if (conn->NextToken() != Token::RPAREN)
    throw Response::BAD(&Msg::ExpectedRParen);
}

static String eat_string(Conn *conn) {
  switch (conn->NextToken()) {
  case Token::QSTRING:
    return conn->Lexer().Quoted();
  case Token::LITERAL:
    return conn->Lexer().Literal();
  default:
    throw Response::BAD(&Msg::ExpectedString);
  }
}

static Ptr<String> eat_string_or_nil(Conn *conn) {
  switch (conn->NextToken()) {
  case Token::QSTRING:
    return new String(conn->Lexer().Quoted());
  case Token::LITERAL:
    return new String(conn->Lexer().Literal());
  case Token::A_NIL:
    return NULL;
  default:
    throw Response::BAD(&Msg::ExpectedStringOrNil);
  }
}

static String eat_qstring(Conn *conn) {
  switch (conn->NextToken()) {
  case Token::QSTRING:
    return conn->Lexer().Quoted();
  case Token::LITERAL:
    return conn->Lexer().Literal();
  default:
    throw Response::BAD(&Msg::ExpectedQString);
  }
}

static unsigned long eat_number(Conn *conn) {
  if (conn->NextToken() == Token::NUMBER) {
    errno = 0;
    // Yeah, we're happily fielding 64bit numbers if we have 64bit longs,
    // but I don't really see that as a big deal (after all, the spec
    // doesn't say that the server MUST reject large numbers... :-)
    unsigned long result = strtoul(conn->Lexer().YYText(), NULL, 10);
    if (errno)
      throw Response::BAD(&Msg::NumberTooBig);
    return result;
  } else
    throw Response::BAD(&Msg::ExpectedNumber);
}

static void eat_store_mods(Conn *conn, StoreEntry *entry) {
  while(1)
    switch (conn->NextToken()) {
    case Token::A_UNCHANGEDSINCE:
      eat_space(conn);
      entry->SetUnchangedSince(eat_string(conn));
      eat_space(conn);
      break;
    case Token::A_NOCREATE:
      entry->SetNoCreate();
      eat_space(conn);
      break;
    default:
      conn->PutbackToken();
      return;
  }
}

static void eat_metadata_string_list(Conn *conn, StoreEntry::Value *value) {
  while(1) {
    value->GetValues().push_back(eat_string(conn));
    switch (conn->NextToken()) {
    case Token::RPAREN:
      return;
    case Token::SPACE:
      break;
    default:
      throw Response::BAD(&Msg::ExpectedRParenOrSpace);
    }
  }
}

static void eat_metadata_write_q(Conn *conn, StoreEntry::Attribute *attr) {

  Token::token metatype = Token::Lookup(eat_string(conn));

  switch (metatype) {
  case Token::A_VALUE:
  case Token::A_ACL:
    break;
  case Token::A_CONTENTTYPE:
    if (attr->Special() == Token::A_ENTRY)
      throw Response::BAD(&Msg::CTypeInEntryAttr);
    break;
  default:
    throw Response::BAD(&Msg::UnrecWriteMetadata);
  }

  eat_space(conn);
  Ptr<StoreEntry::Value> value;
  switch (conn->NextToken()) {
  case Token::QSTRING:
  case Token::LITERAL:
    if (metatype == Token::A_ACL)
      throw Response::BAD(&Msg::ACLString);
    conn->PutbackToken();
    value = new StoreEntry::Value(StoreEntry::Value::STRING);
    value->GetValues().push_back(eat_string(conn));
    break;
  case Token::A_NIL:
    value = new StoreEntry::Value(StoreEntry::Value::NIL);
    break;
  case Token::A_DEFAULT:
    value = new StoreEntry::Value(StoreEntry::Value::DEFAULT);
    break;
  case Token::LPAREN:
    if (metatype == Token::A_CONTENTTYPE)
      throw Response::BAD(&Msg::CTypeStrList);
    if (attr->Special() == Token::A_ENTRY)
      throw Response::BAD(&Msg::NoMultiInEntry);
    value = new StoreEntry::Value(StoreEntry::Value::STRINGLIST);
    eat_metadata_string_list(conn, value);
    break;
  default:
    throw Response::BAD(&Msg::ExpectedValue);
  }
  attr->AddValue(metatype, value);
}

static void eat_store_metadata(Conn *conn, StoreEntry::Attribute *attr) {
  while(1) {
    eat_metadata_write_q(conn, attr);
    switch (conn->NextToken()) {
    case Token::RPAREN:
      return;
    case Token::SPACE:
      break;			// keep looping
    default:
      throw Response::BAD(&Msg::ExpectedStoreMetadata);
    }
  }
}

static void eat_attribute_stores(Conn *conn, StoreEntry *entry) {
  if (conn->NextToken() == Token::RPAREN)
    return;
  else
    conn->PutbackToken();

  while(1) {
    switch(conn->NextToken()) {
    case Token::QSTRING:
    case Token::LITERAL:
      {
	conn->PutbackToken();
	String name = eat_string(conn);
	Token::token special = Token::Lookup(name);
	switch (special) {
	case Token::A_ENTRY:
	  break;
	case Token::A_MODTIME:
	  throw Response::BAD("The modtime attribute is read-only");
	case Token::A_SUBDATASET:
	  break;
	case Token::A_DATASETINHERIT:
	  break;
	default:
	  special = Token::GARBAGE;
	  String::const_iterator it;
	  for (it = name.begin();
	       it != name.end();
	       ++it)
	    if (*it == '.') break;
	  if (it == name.end())
	    throw Response::BAD("Attribute names not containing '.'s are reserved");
	}
	Ptr<StoreEntry::Attribute> attr = new StoreEntry::Attribute(name, special);
	eat_space(conn);
	switch(conn->NextToken()) {
	case Token::A_NIL:
	  attr->SetValueNil();
	  break;
	case Token::A_DEFAULT:
	  attr->SetValueDefault();
	  break;
	case Token::QSTRING:
	case Token::LITERAL:
	  {
	    conn->PutbackToken();
	    Ptr<StoreEntry::Value> value
	      = new StoreEntry::Value(StoreEntry::Value::STRING);
	    value->GetValues().push_back(eat_string(conn));
	    attr->AddValue(Token::A_VALUE, value);
	    break;
	  }
	case Token::LPAREN:
	  eat_store_metadata(conn, attr);
	  break;
	default:
	  throw Response::BAD("Expected value or metadata list");
	}
	if (attr->Special() == Token::A_SUBDATASET) {
	  // We need to validate the subdataset attribute
	  // If it has a value, the value has to be one of:
	  //    DEFAULT,
	  //    NIL,
	  // 	a single value consisting of a '.'
	  // or a single or multivalue, all of which are valid ACAP URLs.
	  Value *value = attr->GetValue();
	  if (value)
	    switch (value->GetValueType()) {
	    case StoreEntry::Value::NIL:
	    case StoreEntry::Value::DEFAULT:
	      break;		// these are okay
	    case StoreEntry::Value::STRING:
	      // Must be either '.' or an ACAP URL.
	      if (value->GetValues()[0] != "."
		  && ! url_dfa.Accepts(value->GetValues()[0]))
		throw Response::Invalid("subdataset", &Msg::InvalidSubdatasetSyntax);
	      break;
	    case StoreEntry::Value::STRINGLIST:
	      // All must be ACAP URLs, and there must be at least one.
	      if (value->GetValues().empty())
		throw Response::Invalid("subdataset", &Msg::InvalidSubdatasetSyntax);
	      for (StoreEntry::Value::values_t::const_iterator it = value->GetValues().begin();
		   it != value->GetValues().end();
		   ++it)
		if (! url_dfa.Accepts(*it))
		  throw Response::Invalid("subdataset", &Msg::InvalidSubdatasetSyntax);
	      break;
	    }
	}
	if (attr->Special() == Token::A_DATASETINHERIT
	    && attr->GetValue()
	    && attr->GetValue()->GetValueType() == StoreEntry::Value::STRINGLIST)
	  throw Response::Invalid("dataset.inherit", &Msg::MultivaluedInherit);

	entry->AddAttribute(attr);

	switch(conn->NextToken()) {
	case Token::RPAREN:
	  return;
	case Token::SPACE:
	  break;
	default:
	  throw Response::BAD("Expected either a space or a ')' after an attribute value");
	}

	break;
      }
    default:
      throw Response::BAD("Expected attribute");
    }
  }
}

static void eat_store_entry(Conn *conn, Store *store) {
  eat_lparen(conn);
  Ptr<StoreEntry> entry = new StoreEntry(eat_string(conn));
  eat_space(conn);
  eat_store_mods(conn, entry);
  eat_attribute_stores(conn, entry);
  store->AddEntry(entry);
}

static void eat_store_entry_list(Conn *conn, Store *store) {
  eat_space(conn);
  eat_store_entry(conn, store);
  while(1)
    switch (conn->NextToken()) {
    case Token::SPACE:
      eat_store_entry(conn, store);
      break;
    case Token::CRLF:
      return;
    default:
      throw Response::BAD("Expected a store-entry or a CRLF");
    }
}

static void eat_search_makecontext(Conn *conn, Modifiers *mods) {
  while(1) {
    eat_space(conn);
    switch (conn->NextToken()) {
    case Token::A_ENUMERATE:
      mods->SetEnumerate();
      break;
    case Token::A_NOTIFY:
      mods->SetNotify();
      break;
    default:
      conn->PutbackToken();
      mods->SetMakeContext(eat_string(conn));
      return;
    }
  }
}

static void eat_search_return_list(Conn *conn, Modifiers *mods) {
  mods->SettingReturnList();
  eat_lparen(conn);
  switch (conn->NextToken()) {
  case Token::RPAREN:
    mods->DoneWithReturnList();
    return;
  default:
    conn->PutbackToken();
    while(1) {
      String attribute = eat_string(conn);
      Ptr<Modifiers::Metadata> metadata = new Modifiers::Metadata();

      if (conn->NextToken() == Token::LPAREN) {
	metadata->AddMetadatum(eat_qstring(conn));
	while(1) {
	  if (conn->NextToken() == Token::RPAREN)
	    break;
	  else {
	    conn->PutbackToken();
	    eat_space(conn);
	    metadata->AddMetadatum(eat_qstring(conn));
	  }
	}
      } else {
	conn->PutbackToken();
	metadata->UseDefaultMetadata(attribute);
      }

      mods->AddReturn(attribute, metadata);

      switch (conn->NextToken()) {
      case Token::SPACE:
	continue;
      case Token::RPAREN:
	mods->DoneWithReturnList();
	return;
      default:
	throw Response::BAD("Expected an attribute or a ')'");
      }
    }
  }
}

static void eat_search_sort_list(Conn *conn, Modifiers *mods) {
  mods->SettingSortList();
  eat_lparen(conn);
  mods->AddSort(eat_string(conn), (eat_space(conn), eat_string(conn)));
  while(1)
    switch(conn->NextToken()) {
    case Token::RPAREN:
      mods->DoneWithSortList();
      return;
    case Token::SPACE:
      mods->AddSort(eat_string(conn), (eat_space(conn), eat_string(conn)));
      continue;
    default:
      throw Response::BAD("Expected a space or a ')'");
    }
}


static void eat_search_modifiers(Conn *conn, Modifiers *mods) {
  while(1) {
    eat_space(conn);
    switch (conn->NextToken()) {
    case Token::A_DEPTH:
      eat_space(conn);
      mods->SetDepth(eat_number(conn));
      break;
    case Token::A_HARDLIMIT:
      eat_space(conn);
      mods->SetHardLimit(eat_number(conn));
      break;
    case Token::A_LIMIT:
      eat_space(conn);
      mods->SetLimit(eat_number(conn), eat_number(conn));
      break;
    case Token::A_MAKECONTEXT:
      eat_search_makecontext(conn, mods);
      break;
    case Token::A_NOINHERIT:
      mods->SetNoInherit();
      break;
    case Token::A_RETURN:
      eat_space(conn);
      eat_search_return_list(conn, mods);
      break;
    case Token::A_SORT:
      eat_space(conn);
      eat_search_sort_list(conn, mods);
      break;
    default:
      conn->PutbackToken();
      return;
    }
  }
}

static Ptr<Criteria> eat_search_criteria(Conn *conn, bool *got_range) {
  switch (conn->NextToken()) {
  case Token::A_ALL:
    return new CritALL();
  case Token::A_AND:
    {
      eat_space(conn);
      Ptr<Criteria> first = eat_search_criteria(conn, got_range);
      eat_space(conn);
      Ptr<Criteria> second = eat_search_criteria(conn, got_range);
      return new CritAND(first, second);
    }
  case Token::A_COMPARE:
    return new CritCOMPARE((eat_space(conn), eat_string(conn)),
			   (eat_space(conn), eat_string(conn)),
			   (eat_space(conn), eat_string(conn)));
  case Token::A_COMPARESTRICT:
    return new CritCOMPARESTRICT((eat_space(conn), eat_string(conn)),
				 (eat_space(conn), eat_string(conn)),
				 (eat_space(conn), eat_string(conn)));
  case Token::A_EQUAL:
    return new CritEQUAL((eat_space(conn), eat_string(conn)),
			 (eat_space(conn), eat_string(conn)),
			 (eat_space(conn), eat_string_or_nil(conn)));
  case Token::A_NOT:
    {
      eat_space(conn);
      Ptr<Criteria> clause = eat_search_criteria(conn, got_range);
      return new CritNOT(clause);
    }
  case Token::A_OR:
    {
      eat_space(conn);
      Ptr<Criteria> first = eat_search_criteria(conn, got_range);
      eat_space(conn);
      Ptr<Criteria> second = eat_search_criteria(conn, got_range);
      return new CritOR(first, second);
    }
  case Token::A_PREFIX:
    return new CritPREFIX((eat_space(conn), eat_string(conn)),
			  (eat_space(conn), eat_string(conn)),
			  (eat_space(conn), eat_string(conn)));
  case Token::A_RANGE:
    *got_range = true;
    return new CritRANGE((eat_space(conn), eat_number(conn)),
			 (eat_space(conn), eat_number(conn)),
			 (eat_space(conn), eat_string(conn)));
  case Token::A_SUBSTRING:
    return new CritSUBSTRING((eat_space(conn), eat_string(conn)),
			     (eat_space(conn), eat_string(conn)),
			     (eat_space(conn), eat_string(conn)));
  default:
    throw Response::BAD("Expected a search criteria");
  }
}

static void eat_command(Conn *conn, String &tag) {
  Ptr<Task> command;

  // dispatch_grab is used to grab the dispatch queue while a command
  // grabs a timestamp and gets whatever other resources need to be
  // acquired; it's there to make sure there's a strict ordering to
  // when commands get threads.  Without this, it'd be possible for
  // every processing thread to be a search, all blocked on some store
  // that's waiting to acquire a thread; by ordering like this, we
  // ensure that the store will get into the queue before any
  // subsequent searches.
  Ptr<Grab> dispatch_grab;

  switch (conn->NextToken()) {
  case Token::A_AUTHENTICATE:
    {
      Grab g(conn->AuthMutex());
      if (conn->IsAuthentic())
	throw Response::BAD("Inappropriate state for command; you're already authenticated");
      DEBUG("Getting a space");
      eat_space(conn);
      DEBUG("Getting auth mech");
      String mech = eat_string(conn);
      DEBUG("Got mech " << mech);
      mech.toLower();
      Auth::MechMap::iterator it = Auth::Mechs().find(mech);
      if (it == Auth::Mechs().end())
	throw Response::BAD("Unknown authentication mechanism");
      (*it).second(conn);
      throw Response::OK(&Msg::Authenticate);
    }
  case Token::A_FREECONTEXT:
    {
      eat_space(conn);
      String ctx = eat_string(conn);
      eat_crlf(conn);
      Grab g(conn->ContextsMutex());
      if (! conn->Contexts().erase(ctx))
	throw Response::BAD(&Msg::NoSuchContext);
      Grab pg(conn->PeerContextsMutex());
      conn->NotifyPeers(ctx);
      throw Response::OK(&Msg::FreeContext);
    }
  case Token::A_LANG:
    {
      const Language *language = NULL;
      while(1) {
	if (conn->NextToken() == Token::CRLF)
	  break;
	else {
	  conn->PutbackToken();
	  eat_space(conn);
	  language = Language::Find(eat_qstring(conn));
	  if (language) {
	    while(conn->NextToken() != Token::CRLF) {
	      conn->PutbackToken();
	      eat_space(conn);
	      eat_qstring(conn);
	    }
	    break;
	  }
	}
      }
      if (! language)
	throw Response::NO("Unable to match any of the specified languages");
      conn->SetLanguage(language);
      {
	Grab g(conn->OutputMutex());
	Compare::comparatorset_t::const_iterator it;
	Compare::comparatorset_t *comp
	  = &(Compare::Instance()->FindFamily("i")->Comparators());
	conn->Out() << tag << " Lang " << &(language->Name());
	for (it = comp->begin(); it != comp->end(); ++it)
	  conn->Out() << ' ' << &(*it);
	comp = &(language->Family()->Comparators());
	for (it = comp->begin(); it != comp->end(); ++it)
	  conn->Out() << ' ' << &(*it);
	conn->Out() << "\r\n";
      }
      throw Response::OK(&Msg::Lang);
    }
  case Token::A_LOGOUT:
    {
      eat_crlf(conn);
      Grab g(conn->OutputMutex());
      conn->Out() << "* Bye \""
		  << conn->CurrentLanguage()->Messages()->Bye() << "\"\r\n"
	// XXX Make this language-neutral...
		  << tag << " Ok \"Logout complete\"\r\n";
      throw Conn::Logout();
    }
#ifndef NDEBUG
  case Token::A_QUIT:
    eat_crlf(conn);
    ACAP::Instance()->Shutdown();
#endif
  case Token::A_DELETEDSINCE:
#ifndef EUDORA_HACKS
    {
      Grab g(conn->AuthMutex());
      if (! conn->IsAuthentic())
	throw Response::BAD("Inappropriate state for command; authenticate first");
    }
#endif // EUDORA_HACKS
    {
      eat_space(conn);
      DeletedSince *deletedsince = new DeletedSince(tag, conn,
						    eat_string(conn),
						    (eat_space(conn),
						     eat_string(conn)));
      command = deletedsince;
      dispatch_grab = new Grab(Dispatch::Instance()->OrderingMutex());
      deletedsince->Setup();
      eat_crlf(conn);
      break;
    }
  case Token::A_NOOP:
    eat_crlf(conn);
    throw Response::OK(&Msg::Noop);
  case Token::A_SEARCH:
    {
#ifndef EUDORA_HACKS
      {
	Grab g(conn->AuthMutex());
	if (! conn->IsAuthentic())
	  throw Response::BAD("Inappropriate state for command; authenticate first");
      }
#endif // EUDORA_HACKS
      eat_space(conn);
      Search *search = new Search(tag, conn, eat_string(conn));
      command = search;
      eat_search_modifiers(conn, search->GetModifiers());
      bool got_range = false;
      Ptr<Criteria> criteria = eat_search_criteria(conn, &got_range);
      search->SetCriteria(criteria);
      eat_crlf(conn);
      dispatch_grab = new Grab(Dispatch::Instance()->OrderingMutex());
      search->Setup();
      if (got_range && ! search->IsEnumeratedContext())
	throw Response::BAD("The Range operator is only valid when searching an enumerated context");
      break;
    }
  case Token::A_STORE:
    {
#ifndef EUDORA_HACKS
      {
	Grab g(conn->AuthMutex());
	if (! conn->IsAuthentic())
	  throw Response::BAD("Inappropriate state for command; authenticate first");
      }
#endif // EUDORA_HACKS
      Store *store = new Store(tag, conn);
      command = store;
      eat_store_entry_list(conn, store);
      dispatch_grab = new Grab(Dispatch::Instance()->OrderingMutex());
      store->Setup();
      break;
    }
  case Token::A_UPDATECONTEXT:
    {
      Grab g(conn->ContextsMutex());
      UpdateContext::contexts_t contexts;
      while(conn->NextToken() != Token::CRLF) {
	conn->PutbackToken();
	eat_space(conn);
	String ctx = eat_string(conn);
	Conn::contexts_t::iterator it = conn->Contexts().find(ctx);
	if (it == conn->Contexts().end())
	  throw Response::NO(&Msg::NoSuchContext);
	NotifyContext *nc = (*it).second->NotifyCast();
	if (! nc)
	  throw Response::NO(&Msg::NotUpdateableContext);
	contexts.push_front(nc);
      }
      if (contexts.empty())
	throw Response::BAD(&Msg::ExpectedContext);
      dispatch_grab = new Grab(Dispatch::Instance()->OrderingMutex());
      command = new UpdateContext(tag, conn, contexts);
      break;
    }
  default:
    throw Response::BAD("Unknown command");
  }

  Dispatch::Instance()->Enqueue(command);
}

void Conn::Loop() throw () {
  try {
    while(1) {
      try {
	// Get a tag...
	lexer.ExpectingTag();
	switch (NextToken()) {
	case Token::TAG:
	  {
	    String tag = String::Copy(lexer.YYText());
	    try {
	      eat_space(this);
	      eat_command(this, tag);
	    } catch (Response::StatusResponse &c) {
	      Grab g(OutputMutex());
	      out << tag << c << flush;
	      Resync();
	    }
	  }
	  break;
	default:
	  throw Response::BAD("Syntax error: expected a tag");
	}
      } catch (Response::StatusResponse &c) {
	Grab g(OutputMutex());
	out << "*" << c << flush;
	Resync();
      }
    }
  } catch (ConnectionClosed) {
    Grab g(OutputMutex());
    out << "* Bye \"You could at least say `logout'\"\r\n";
  } catch (Logout) {
  } catch (Error e) {
    Grab g(OutputMutex());
    out << "* Bye \"" << e.err() << "\"\r\n";
  } catch (...) {
    Grab g(OutputMutex());
    out << "* Bye \"Internal server error: an unexpected exception was thrown, which puts the connection into a non-deterministic state.  We can't really recover, so we have to close it.  Sorry...\"\r\n";
  }
  try {
    DEBUG("Conn: erasing contexts...");
    Grab g(ContextsMutex());
    contexts.erase(contexts.begin(), contexts.end());
    DEBUG("Conn: erased");
  } catch (...) {
  }
  DEBUG("Conn: waiting for references to go away");
}

void Conn::GetCRLF() {
  eat_crlf(this);
}

void Conn::GetSpace() {
  eat_space(this);
}

String Conn::GetNextString() {
  if (NextToken() == Token::STAR)
    throw Response::NO(Msg::CommandAborted);
  PutbackToken();
  return eat_string(this);
}

void Conn::AddPeer(const String &path, NotifyContext *context) {
  peer_contexts[path].insert(context);
}

void Conn::RemovePeer(const String &path, NotifyContext *context) {
  peer_contexts_t::iterator it = peer_contexts.find(path);
  if (it != peer_contexts.end()) {
    (*it).second.erase(context);
    if ((*it).second.empty())
      peer_contexts.erase(it);
  }
}

void Conn::NotifyPeers(const String &path) {
  peer_contexts_t::iterator it = peer_contexts.find(path);
  if (it != peer_contexts.end())
    for (peer_context_t::iterator cit = (*it).second.begin();
	 cit != (*it).second.end();
	 ++cit)
      (*cit)->ScheduleUpdate();
}
