/* entry.h -*- C++ -*- entry interface
 * Rob Earhart
 * $Id: entry.h,v 1.7 1998/03/09 00:58:18 rob Exp $
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_ENTRY_H
#define ACAP_ENTRY_H

#include <hash_map.h>
#include "ptr.h"
#include "str.h"
#include "dsobjects.h"
#include "attribute.h"
#include "dstring.h"

class Dataset;

class Entry {
public:
  Entry(Dataset *_dataset, entryval_t *_data)
    : dataset(_dataset), data(_data) {}
  Entry() : dataset(NULL), data(NULL) {}
  Entry(const Entry &it) : dataset(it.dataset), data(it.data) {}

  Entry &operator=(const Entry &it) {
    dataset = it.dataset;
    data = it.data;
    return *this;
  }

  Timestamp::Stamp CreateTime() const {
    return data->createtime;
  }
  Timestamp::Stamp ModTime() const {
    return data->modtime;
  }
  Timestamp::Stamp DeleteTime() const {
    return data->deletetime;
  }
  bool InDate(Timestamp::Stamp stamp) const {
    return data->in_date(stamp);
  }
  bool Exist() const {
    return data != NULL;
  }
  bool Visible() const {
    return data->attrs != NULL && 2 < data->attrs->n_alloced;
  }
  bool Inherit() const {
    return data->inherit;
  }
  Attribute Lookup(const String &name, Timestamp::Stamp stamp) const;

  typedef hash_map<DString, Attribute, hash<DString>, equal_to<DString> > attributes_t;
  typedef hash_set<DString, hash<DString>, equal_to<DString> > attr_filter_t;

  void Match(attributes_t &result,
	     const String &pattern,
	     Timestamp::Stamp stamp,
	     Ptr<attr_filter_t> &prev_filter) const;

  DString Name() const {
    return DString(&data->name);
  }
  Attribute Subdataset(Timestamp::Stamp stamp) const {
    return Lookup("", stamp);
  }

  bool HasADefaultAttribute(Timestamp::Stamp stamp) const {
    attrlist_t *attrs = data->attrs;
    if (! attrs) return false;
    for (unsigned lupe = 0;
	 lupe < attrs->n_alloced;
	 lupe++) {
      Attribute attr(dataset, attrs->buckets(lupe));
      if (attr.Exist() && attr.InDate(stamp) && attr.IsDefault())
	return true;
    }
    return false;
  }

  // Writing
  void AssignAttributes(Entry &it, Trx *trx) {
    DEBUG("Assigning attributes from entry at " << it.data << " to " << data);
    it.data->set_freeattrs(false, trx);
    DEBUG("Assigning attrs to " << it.data->attrs);
    data->set_attrs(it.data->attrs, trx, (PS *)dataset);
  }
  void SetCreateTime(Timestamp::Stamp stamp,
		     Trx *trx) {
    data->set_createtime(stamp, trx);
  }
  void SetModTime(Timestamp::Stamp stamp,
		  Trx *trx) {
    data->set_modtime(stamp, trx);
    Attribute ma = Lookup("modtime", trx->TrxStamp());
    if (ma.Exist()) ma.SetDelete(trx);
    ma = CreateAttribute("modtime", trx);
    ma.SetSingleValue(trx, String(trx->TrxTimeStr()));
  }
  void SetDeleteTime(Timestamp::Stamp stamp,
		     Trx *trx) {
    data->set_deletetime(stamp, trx);
  }
  void SetInherit(bool inherit,
		  Trx *trx) {
    data->set_inherit(inherit, trx);
  }
  Attribute CreateAttribute(const String &attribute,
			    Trx *trx);
  Dataset *GetDataset() const { return dataset; }

  // This is only here so that Dataset::InheritEntry can make better
  // decisions as to which entries it should copy...
  bool IsSameData(entryval_t *_data) const { return data == _data; }
private:
  Dataset *dataset;
  entryval_t *data;
};

#endif // ACAP_ENTRY_H
