/* entryref.h -*- C++ -*- entry reference interface
 * Rob Earhart
 * $Id: entryref.h,v 1.2 1998/03/09 00:58:20 rob Exp $
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_ENTRYREF_H
#define ACAP_ENTRYREF_H

// Okay.  The basic problem: We really want shared reference-counted
// entries, so that changes show up as appropriate, so that we share
// info when possible, so that the cache stuff is consistant, etc.
// But we also want per-entry per-command (search, context, whatever)
// information.  Thus, we need EntryRef, a class which is itself not
// reference counted, maintains some per-command information, but
// keeps a reference to an entry, which is ultimately used as a data
// source.

#include <vector.h>
#include <algo.h>
#include "entry.h"
#include "str.h"
#include "timestamp.h"
#include "modifiers.h"

class EntryRef {
public:
  typedef Entry::attributes_t attributes_t;

  EntryRef(const Entry &first, Timestamp::Stamp _stamp) : stamp(_stamp), ent_number(0) {
    ents.push_back(first);
  }
  EntryRef(const EntryRef &it)
    : stamp(it.stamp), ent_number(it.ent_number), relpath(it.relpath), ents(it.ents) {}

  void AddInheritedEntry(const Entry &ent) {
    ents.push_back(ent);
  }

  DString Name() const { return ents[0].Name(); }
  Timestamp::Stamp Stamp() const;
  Timestamp::Stamp ModTime() const;
  bool Defaulted() const;
  Attribute Lookup(const String &name) const;
  void Match(attributes_t &result, const String &pattern) const;
  bool IsVisible() const;
  Attribute Subdataset() const {
    return Lookup("");
  }

  void SetEntryNumber(unsigned val) { ent_number = val; }
  unsigned EntryNumber() const { return ent_number; }
  void Write(ostream &stream,
	     Modifiers::returnlist_t::const_iterator start,
	     Modifiers::returnlist_t::const_iterator finish) const;
  void SetRelPath(const String &p) { relpath = p; }
  const String &RelPath() const { return relpath; }

private:
  struct entry_cmp : public binary_function<Entry, Entry, bool > {
    bool operator()(const Entry &a, const Entry &b) const {
      return (a.GetDataset() == b.GetDataset()
	      && a.CreateTime() == b.CreateTime());
    }
  };

public:
  bool operator==(const EntryRef &it) const {
    // This should suffice... since if the inheritance path has
    // changed or if things have changed in interesting ways or if the
    // entryrefs just plain point to completely different things,
    // we'll get the correct result.
    return ents.size() == it.ents.size()
      && equal(ents.begin(), ents.end(),
	       it.ents.begin(),
	       entry_cmp())
      && RelPath() == it.RelPath();
  }

private:
  Timestamp::Stamp stamp;
  unsigned ent_number;
  String relpath;
  typedef vector<Entry> ents_t;
  ents_t ents;			// inheritance chain
};

#endif // ACAP_ENTRYREF_H
