/* interlock.h -*- C++ -*- dataset interlock interface
 * Rob Earhart
 * $Id: interlock.h,v 1.8 1998/03/09 00:58:29 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_INTERLOCK_H
#define ACAP_INTERLOCK_H

#include <pthread.h>
#include <deque.h>
#include <map.h>
#include <set.h>
#include <hash_map.h>
#include "refcount.h"
#include "ptr.h"
#include "str.h"
#include "mutex.h"
#include "cond.h"
#include "timestamp.h"

class Version;

// A DeletedSince and a Search need to be able to, for any piece of
// data in the system, guarantee that no earlier command is still in
// progress.

// A Store needs to be able to look up information, and atomically
// update the values of multiple pieces of data.

// That's really all there is to it.

// So: Each Store operation needs to queue it's change for each piece
// of data which it wants to change (which *should* be known apriori).
// Multiple Stores may be in progress at once, and do NOT have to wait
// for each other to complete, but if they need to read a piece of
// data being set by a prior Store they need to wait for the prior
// Store to complete, and if they need to write a piece of data that
// may be written by a subsequent Store, they need to ensure that they
// do not overwrite the data from the subsequent Store (in case of
// reordering).

// One important optimization: Stores will often not have the NOCREATE
// flag set, and will thus create all intermediate datasets, if they
// do not exist.  This affects the behavior of subsequent operations
// along the path; nevertheless, it differs from the normal behavior
// in that such a Store will not do anything to a path if it already
// exists, so subsequent traversers need not wait for the Store to
// complete in order to traverse iff the object being traversed
// exists.  If it doens't exist, of course, one must wait for the
// Store to either create it or not before knowing whether and how it
// may be traversed.  This must be a seperate check, because it's also
// possible for a Store to modify a piece of data at a path
// irregardless of whether it already exists, and in this case
// subsequent operations *must* wait for the Store to complete.

// Note that DEADLOCK IS NOT POSSIBLE in this scheme, as all
// operations have a unique timestamp and are thus ordered, and early
// operations NEVER block for later operations.  In the worst case,
// with a maximum of data dependancies between operations, the
// operations will complete in the order in which they acquired
// timestamps.

// One important note: Stores MUST atomically acquire their timestamps
// and enter their changes into the Interlock structure, in order to
// ensure that subsequent operations will see that a Store command is
// in progress on those piece of data.

// Note that this synchronization is by name, not by value; a rename
// operation does *not* rename queues, since reads and writes apply to
// datasets by name and not to some sort of virtual identifier that
// remains constant across a dataset's life as it's renamed and
// otherwise modified.  Neither do these map to the dataset data,
// since the name->data mapping is only instantaneously valid; after
// achieving access to a given dataset by name, the thread must then
// go in and seperately synchronize access to the underlying data (in
// case of a write/rename/write sequence).

// Another note: a thread performing a rename operation *does* have to
// lock the destination dataset name, to prevent the dataset from
// being accessed via the new name before it's actually been renamed.

// Fun, eh?  (JKMN)

// For now, we're just going to lock entire entries; someday (XXX) we
// should consider locking just the attributes of the entry that
// we're planning to change.

// A side effect of this is that we only ever lock one kind of thing,
// which makes life a little simpler (even if we have to be able to
// lock it in two ways, for implicit creation and real updates).

class Interlock {
  // The Interlock is a Singleton class responsible for synchronizing
  // the actions of threads upon the underlying ACAP datastore.
public:
  class Version;

  class Resource : public Refcounted {
    // A Resource keeps track of a given piece of named data in the
    // datastore.
  public:
    Resource(const String &name);
    void Noref();		// needed because we need to grab the
				// interlock mutex before deciding whether
				// or not we *really* have no outstanding references
    ~Resource();

    // Wait is used to wait for a resource to become available as of a
    // given timestamp.  It doesn't assume anything at all about the
    // Interlock mutex.
    void Wait(Timestamp::Stamp stamp, bool exists);

    // These two are used by Version objects to add and remove
    // themselves from the lists of versions with various types of
    // locks on this resource.  Neither assumes that the Interlock
    // mutex has been grabbed (neither cares).
    void RemoveVersion(Version *v);
    void AddVersion(Version *v, bool implicit);

    // For convenience...
    const String &Name() { return name; }

  private:
    String name;
    Mutex interlock_mutex;
    Mutex locks_mutex;
    typedef map<Timestamp::Stamp, Ptr<Cond>, less<Timestamp::Stamp> > locks_t;
    locks_t locks;
    locks_t implicit;

    void Wait(Timestamp::Stamp stamp, locks_t &l);

    Timestamp::Stamp modtime;
  };

  class Version {
    // A Version is used by a Store command to keep track of a number
    // of Resource objects, keeping their queues consistant and making
    // sure that commands depending on those resources are notified
    // when they become available (either due to commit or rollback).
  public:
    // Note: Versions should be created only when you've grabbed the
    // Interlock mutex, since you need to get a timestamp and register
    // with all the appropriate resources as an atomic action.

    Version();
    ~Version();

    // Get a version's timestamp
    Timestamp *GetTimestamp() { return timestamp; }

    // AddResource indicates that this version will modify the
    // specified resource; implicit_create indicates that the resource
    // will be created iff it doesn't already exist.
    // It assumes that the Interlock mutex is grabbed.
    void AddResource(const String &name, bool implicit_create);

  private:
    Ptr<Timestamp> timestamp;
    typedef set<P<Resource>, less<Resource *> > resources_t;
    resources_t resources;
  };

  // Used to access the Interlock object
  static Interlock *Instance();

  // This is used to create a timestamp by commands which don't need
  // Version objects; it does so within a grab of the Interlock,
  // making sure that any prior stores have registered themselves, so
  // that the generated timestamp is suitable for traversal checking.
  // It assumes that the Interlock mutex is not grabbed.
  Ptr<Timestamp> CreateStamp() {
    Grab g(Mutex());
    return TimestampFactory::Instance()->NewStamp();
  }

  // ... and this is what's actually used for said traversal checking.
  // "exists" should be set iff the resource is known to exist; if
  // set, it won't block just because something's promising to create
  // it iff the resource doesn't exist (if not set, it must block for
  // these, since it may or may not be created).
  // It assumes that the Interlock mutex is not grabbed.
  void WaitForResource(const String &name,
		       Timestamp::Stamp stamp,
		       bool exists);

  // This is used by Versions to look up resources.  It
  // assumes that the Interlock mutex is grabbed.
  P<Resource> LookupResource(const String &name);
  
  // This is used by Resources to unlink themselves when they're no
  // longer referenced by anything.
  // It assumes that the Interlock mutex is grabbed.
  void RemoveResource(const String &name);

  // Used for synchronized access to the interlock
  Mutex &InterlockMutex() { return interlock_mutex; }

private:
  Interlock() {}
  static void Initialize();
  static Interlock *instance;
  static pthread_once_t instance_init;

  Mutex interlock_mutex;
  typedef hash_map<const String, Resource *,
                   hash<String>, equal_to<String> > resources_t;
  resources_t resources;
};

#endif // ACAP_INTERLOCK_H
