/* page.h -*- C++ -*- interface to pages
 * Rob Earhart
 * $Id: page.h,v 1.1 1998/02/18 21:19:08 rob Exp $
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_PAGE_H
#define ACAP_PAGE_H

#include <stdio.h>
#include <hash_map.h>
#include <slist.h>
#include "mutex.h"
#include "refcount.h"

// Okay.  Some comments about how all this works:

// (Note: the idea for this comes from the OOPS Research Group at the
// University of Texas <url:http://www.cs.utexas.edu/users/oops/>; I
// looked at incorporating their Texas data store, but it wasn't a
// good fit with what the ACAP server's doing (read: threads), so I
// decided to just write my own persistent storage module using
// pointer swizzling at page-fault time; given the limited scope of
// what I'm trying to do, this seemed like a reasonable approach.)

// Each dataset's data is handled by a single persistent store (PS)
// object, which is responsible for managing the address space
// occupied by the data.  The global address space is managed by a
// singleton object, Pages.

// The initial page in a dataset is mapped in with no access
// permissions, and its address is mapped to a PS object in the Pages
// singleton.  When an attempt is made to access the page, a
// segmentation violation is triggered by the virtual memory hardware,
// which causes the thread to call the static function
// Pages::PageFaultHandler, which finds the singleton Pages instance
// and calls its FaultHandler method.

// Pages::FaultHandler uses the address to determine which PS object
// the page belongs to, and asks that PS to swizzle in the page.

// A note on swizzling: pointers on disk are stored as offsets; they
// have to be, since we have no idea at runtime where a given page of
// a dataset will be mapped.  When we swizzle a page, we translate all
// those offsets into addresses.  If an offset refers to a part of a
// file that's been mapped in (and thus has an entry in the PS's
// offset to address mapping table), we just store the appropriate
// pointer; if there is no mapping for that page of the file, we
// establish one, mapping in a non-accessible page as a placeholder,
// and storing a reference to it in the Pages singleton.  Thus, each
// pointer in a swizzled page points either to another swizzled page
// or to a virtual address which, when used, will cause a segmentation
// violation, which will swizzle in that page, creating virtual
// mappings for all the pages *it* points to... it's actually fairly
// elegant.  You pay a penalty for the first access to a page, but
// it's effectively the same penalty as for faulting in the page
// anyway, and you minimize commitment of your virtual address space.
// For more info, check out that OOPS link above.

// Okay, so that's how we swizzle.  The Pages singleton needs to do a
// bit more work than that, though; the page must be read/writeable to
// swizzle it, but we can't make it read/writeable until it's been
// swizzled, because we need to keep other threads from accessing it.

// So what we do is a little sick: we have a seperate, temporary file
// open (unlinked, so its disk storage will be reclaimed when we
// exit).  When we want to swizzle, we map in a MAP_SHARED temporary
// page of that temporary file at a *different* address, and read the
// page from the PS's file into that address.  We then tell the PS to
// swizzle *that* page - which works just fine, since swizzling just
// walks the pointers and rewrites them, never traversing them, so
// even internal pointers will come out correctly (they'll point at
// the virtual address where we eventually want the page to live).
// Then we unmap that temporary page, and map it MAP_PRIVATE with
// PROT_READ | PROT_WRITE at its final location (which atomically
// turns the page of data into the swizzled page, and sets the
// protections, so we get thread safety), and write to it to make sure
// its backing store is coming from swap and not from our temporary
// file (we want to disassociate it from the file immediately).  We
// can then reuse that offset of our temporary file for another
// swizzling operation, and keep using swap for its intended purpose
// (swap's always at least as fast and efficient as paging to a
// temporary file on disk, and often quite a bit faster and more
// efficient, since most machines are configured s.t. swap doesn't
// have to go through the file system, and swapping can be
// parallelized nicely if it's spread out, etc.)

// We have to use a temporary file because we can't say "mmap this
// page of memory at this address with this protection"; we can only
// mmap offsets of file descriptors.  Thank you, unix.  (Could be
// worse, though).

// When a PS is destroys itself, it simply unmaps all its pages and
// removes them from the address->PS mapping (technically, this last
// bit isn't necessary, since if the pages are unmapped nothing should
// try to use them anyway, but I'd rather have the page fault handler
// be able to determine whether or not their *should* be a mapping at
// a given address, rather than just trying to muddle through...

const unsigned pagebits = 12;	// XXXXXXXXXXXX - maybe this should be a runtime thing?
const off_t pagesize = (1 << pagebits);
const off_t pagesize_m1 = (pagesize - 1);

struct PageOffsetHash {
  size_t operator()(off_t address) const { return address >> pagebits; }
  size_t operator()(caddr_t address) const { return (unsigned)address >> pagebits; }
};

// Pages is a singleton class which keeps track of the allocated pages
// and mappings in the system.
struct PS;

class Pages {
public:
  static Pages *Instance() {
    if (! instance)
      pthread_once(&instance_init, &Initialize);
    return instance;
  }

  Mutex &MapMutex() { return map_mutex; }

  // PageFaultHandler is meant to be called by SIGSEGV
  // to perform swizzling.
  static void PageFaultHandler(int sig, siginfo_t *sip, void *up);

  // Both of these should be called only when the MapMutex is held.
  void EstablishMapping(caddr_t addr, PS *ps, off_t offset) {
    DEBUG("Mapping address " << (void *)addr << " to ps " << ps);
    mappings.insert(mappings_t::value_type(addr, new Page(ps, offset)));
  }
  void RemoveMapping(caddr_t addr) {
    DEBUG("Removing mapping for address " << (void *)addr);
    mappings.erase(addr);
  }

private:
  Pages();
  static void Initialize();
  static Pages *instance;
  static pthread_once_t instance_init;

  // This is what PageFaultHandler really calls
  void FaultHandler(int sig, siginfo_t *sip, void *up);

  FILE *tmp;			// where swizzled pages live temporarily
				// we use a FILE * for this so that it can
				// be initialized by tmpfile(); once created,
				// we only ever use it for the fd.
  size_t tmpsize;		// current size of tmp

  // Page keeps track of whether or not we've swizzled a given page.
  class Page : public Refcounted {
  public:
    Page(PS *_ps, off_t _offset) : ps(_ps), offset(_offset), swizzled(false) {}
    Mutex mutex;
    PS *ps;
    bool swizzled;
    off_t offset;
  };

  typedef hash_map<caddr_t, P<Page>,
                   PageOffsetHash, equal_to<void *> > mappings_t;
  typedef slist<off_t> freeswap_t;
  mappings_t mappings;
  freeswap_t freeswap;
  Mutex map_mutex;
};

#endif // ACAP_PAGE_H
