/* ps.cc - persistent store implementation
 * Rob Earhart
 * $Id: ps.cc,v 1.2 1998/03/09 00:58:41 rob Exp $
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <sys/types.h>
#include "ps.h"
#include "timestamp.h"
#include "unixerror.h"
#include "trx.h"

static inline size_t RoundUp(size_t n, size_t align) {
  return ((n+(align-1))/align)*align;
}

static int ZeroFD() {
  static Mutex m;
  static int fd = -1;

  if (fd < 0) {
    Grab g(m);
    if (fd < 0) {
      fd = open("/dev/zero", O_RDWR);
      if (fd < 0)
	throw UnixError("Unable to access /dev/zero");
    }
  }
  return fd;
}

PSBlob::PSBlob(PS *_ps, void *_addr)
    : ps(_ps), addr(_addr), free_scheduled(false) {}

void PSBlob::ScheduleFree() {
  free_scheduled = true;
}

PSBlob::~PSBlob() {
  if (free_scheduled)
    ps->Recycle(addr);
  ps->RemoveBlob(addr);
}

void PSBlob::Noref() {
  // We play this little game with dependants because, if we're the last
  // reference to these dependants, we *don't* want our delete method
  // to call their Noref method and have them block on the PSMutex
  // which we're holding at a higher point on the call stack.
  // So, if we're going to delete ourselves, we move the dependants
  // over to the depend list, which will be destroyed *after* the Grab
  // goes away.
  dependants_t depend;
  Grab g(ps->PSMutex());
  if (CheckNoRefs()) {
    depend.swap(dependants);
    delete this;
  }
}

PSResources::PSResources(PS *_ps)
  : ps(_ps) {}

PSResources::~PSResources() {
  ps->resources = NULL;
  ps->FlushPages();
}

void PSResources::Noref() {
  Grab g(ps->resources_mutex);
  if (CheckNoRefs())
    delete this;
}

P<PSResources> PS::GetResources() {
    Grab g(resources_mutex);
    P<PSResources> result;
    if (resources) result = resources;
    else resources = result = new PSResources(this);
    return result;
}

PS::object_t::object_t(char tag, Trx *trx) {
  header.tag = tag;
  trx->WriteData(PageOffsetOfAddr(&header),
		 &header,
		 sizeof(header));
}

struct fmem_t {
  unsigned n_buckets;
  struct buckets_t {
    off_t offset;
    size_t size;
  };
  buckets_t buckets[1];
};

void PS::Init(Trx *trx) {
  // XXX This should probably be done as part of the trx
  if (inited) return;

  if (fd < 0) fd.Open(path.c_str(), O_RDWR | O_CREAT);
  if (fstat(fd, &sb))
    throw UnixError("statting datastore file");
      
  if (! sb.st_size) {
    if (! trx)
      throw InternalError("Attempt to read from an uninitialized datastore");
    sb.st_size = pagesize;
    trx->SetPS(this);
    off_t offset;
    root = CreateRoot(trx, &offset);
    assert(offset == pagesize + sizeof(align_t)); // paranoia
    assert(root == AddressOfOffset(pagesize) + sizeof(align_t)); // more paranoia
  } else {
    fmem_t *fmem = (fmem_t *)MapPage(0, map_file, PROT_READ | PROT_WRITE);
    for (unsigned lupe = 0; lupe < fmem->n_buckets; lupe++)
      freemem.insert(freemem_t::value_type(fmem->buckets[lupe].size,
					   fmem->buckets[lupe].offset));
    root = AddressOfOffset(pagesize) + sizeof(align_t);
  }
  inited = true;
}

PS::~PS() {
  Grab g(Pages::Instance()->MapMutex());
  for (offsets_t::const_iterator it = offsets.begin();
       it != offsets.end();
       ++it) {
    Pages::Instance()->RemoveMapping((*it).first);
    munmap((*it).first, pagesize);
  }
}

// XXX Are we *sure* this interacts correctly?  Hrm.
void PS::FlushPages() {
  Grab g(PSMutex());		// Keep anything else from playing with our pages;
				// we don't want to flush pages out from beneath
				// some other thread...
  for (offsets_t::const_iterator it = offsets.begin();
       it != offsets.end();
       ++it)
    // We can't just munmap the page from our address space;
    // we need to make sure the kernel doesn't reuse the address
    // for another page later.  But we can create a dummy mapping
    // for the page fairly easily... all we really need is a PROT_NONE
    // page mapped from somewhere (shared/private doesn't matter, since
    // the page won't be actually allocated any physical memory until we
    // fault on it, at which point it'll be remapped to something else).
    if ((*it).second)
      mmap((*it).first, pagesize, PROT_NONE, MAP_SHARED, fd, 0);
}

caddr_t PS::AddressOfOffset(off_t offset) {
  if (! offset) return NULL;
  off_t base = offset & ~pagesize_m1;
  assert(base <= offset && offset < base + pagesize);
  caddr_t addr;
  addresses_t::const_iterator it = addresses.find(base);
  if (it == addresses.end()) {
    // We need to map in the page
    addr = MapPage(base, map_zero, PROT_NONE);
  } else {
    // We've already mapped in the page
    addr = (*it).second;
  }
  return (caddr_t)((unsigned)addr | offset & pagesize_m1);
}

off_t PS::OffsetOfAddress(void *ptr) {
  if (! ptr) return 0;
  caddr_t base = (caddr_t)((unsigned)ptr & ~pagesize_m1);
  assert(base <= ptr && ptr < base + pagesize);
  offsets_t::const_iterator it = offsets.find(base);
  if (it == offsets.end())	// we should never be asked to unswizzle
    abort();			// an address that we don't have an offset for;
				// if this happens, something is SERIOUSLY
				// wrong...
  return (*it).second | (unsigned)ptr & pagesize_m1;
}

caddr_t PS::MapPage(off_t offset, map_type mt, int prot) {
  caddr_t address;
  if (mt == map_file)
    address = mmap(NULL, pagesize, prot, MAP_PRIVATE, fd, offset);
  else
    address = mmap(NULL, pagesize, prot, MAP_PRIVATE, ZeroFD(), 0);
  //DEBUG("Mapping offset " << (void *)offset << " to address " << (void *)address << " in ps " << this);
  addresses[offset] = address;
  offsets[address] = offset;
  {
    Grab g(Pages::Instance()->MapMutex());
    Pages::Instance()->EstablishMapping(address, this, offset);
  }
  return address;
}

void PS::SwizzlePage(caddr_t realaddr, caddr_t addr) {
  // The page should already be in memory at this point,
  // so everything we're doing should be in memory, so
  // barring swapping, this should be reasonably fast;
  // so it's better to grab the PSMutex once here, instead of
  // grabbing it for each swizzle.
  Grab g(PSMutex());

  // This makes sure we're all kosher with respect to offsets -
  // OffsetOfAddress is the offset the PS thinks the page is at,
  // PageOffsetOfAddr is the offset the page thinks it's at.
  assert(OffsetOfAddress(realaddr) == PageOffsetOfAddr(addr));

  // And then we just walk the page, swizzling everything.
  //DEBUG("Swizzling page at " << (void *)(addr + sizeof(align_t)));
  for (caddr_t thisobj = addr + sizeof(align_t);
       thisobj < addr + pagesize;
       thisobj += ((object_t *)thisobj)->header.size) {
    //DEBUG("Swizzling object at " << (void *)thisobj);
    //DEBUG("Object tag is " << ((object_t *)thisobj)->header.tag);
    SwizzleObject(thisobj, ((object_t *)thisobj)->header.tag);
  }
}

alloc_t PS::Alloc(size_t nbytes, Trx *trx) {
  // Might as well do this check here;
  // if this fails, make sure you *slowly* squeeze the trigger on the
  // gun pointed at the guy who designed your OS; you wouldn't want
  // to jerk and miss.
  assert(! (pagesize % sizeof(align_t)));

  trx->SetPS(this);

  if (nbytes < sizeof(align_t))
    nbytes = sizeof(align_t);
  nbytes = RoundUp(nbytes, sizeof(align_t));
  if (pagesize < nbytes + sizeof(align_t))
    throw InternalError("Large object allocation requested; unimplemented (yet).");
  Grab g(ps_mutex);

  // An ideal implementation would grab the first-fit from the pages
  // that're already swizzled in, and then fallback.  XXX
  off_t offset;
  size_t size;
  caddr_t addr = NULL;

  freemem_t::iterator it = freemem.lower_bound(nbytes);
  if (it == freemem.end()) {
    // Grn - gotta allocate some more memory.
    offset = sb.st_size + sizeof(align_t);
    size = pagesize - sizeof(align_t);

    // We pull it in Now from /dev/zero, instead of from the file
    // (which hasn't actually been grown yet).  And since there's
    // nothing in the page, swizzling it would be pointless, so we
    // just map it in read/write.
    addr = MapPage(sb.st_size, map_zero, PROT_READ | PROT_WRITE);
    *(off_t *)addr = sb.st_size;
    addr += sizeof(align_t);
    trx->WriteGrowTo(sb.st_size + pagesize);
    trx->WriteOffset(offset - sizeof(align_t), sb.st_size);
    sb.st_size += pagesize;
  } else {
    size = (*it).first;
    offset = (*it).second;
    freemem.erase(it);
    trx->WriteAlloc(offset, size);
    addr = AddressOfOffset(offset);
  }

  object_t *baseobj = (object_t *)addr;
  baseobj->header.size = nbytes;
  // The size will be written out when the tag is written out...

  if (nbytes < size) {
    assert(size - nbytes >= sizeof(align_t));
    freemem.insert(freemem_t::value_type(size - nbytes, offset + nbytes));
    trx->WriteFree(offset + nbytes, size - nbytes);
    object_t *newobj = (object_t *)(addr + nbytes);
    newobj->header.size = size - nbytes;
    newobj->header.tag = 0;
    trx->WriteData(PageOffsetOfAddr(&newobj->header),
		   &newobj->header,
		   sizeof(newobj->header));
  }

  // And we're done!
  return alloc_t(addr, offset);
}

P<PSBlob> PS::GetBlob(void *addr) {
  Grab g(ps_mutex);
  P<PSBlob> result;
  blobs_t::iterator it = blobs.find(addr);
  if (it != blobs.end()) result = (*it).second;
  else {
    result = new PSBlob(this, addr);
    blobs[addr] = result;
  }
  return result;
}

P<PSBlob> PS::ScheduleFree(void *addr, Trx *trx) {
  //DEBUG("Scheduling free for " << addr);
  P<PSBlob> blob = GetBlob((caddr_t)addr);
  blob->ScheduleFree();
  trx->AddPSBlob(blob);

  // Hopefully, the page is already in memory...
  object_t *object = (object_t *)addr;
  trx->WriteFree(OffsetOfAddress(addr), object->header.size);
  object_t obj = *object;
  obj.header.tag = 0;		// write the empty_tag back to disk...
  trx->WriteData(PageOffsetOfAddr(&object->header),
		 &obj,
		 sizeof(object->header));
  return blob;
}

void PS::Recycle(void *addr) {
  // We don't have to grab the ps_mutex, because
  // this should only ever be called by an FSBlob
  // that's in the middle of being deleted.

  // XXX Do free block coalescing at some point...
  // Hopefully, the page is already in memory...

  object_t *object = (object_t *)addr;

  // XXX Actually, come to think of it, I'm pretty sure we never have
  // to execute the next line of code: why set the tag?  It'll get set
  // iff the object's reallocated, and if it's not, it's already set
  // on disk, so it'll get swizzled in to the correct value eventually
  // anyway...

  // object->header.tag = empty_tag;

  freemem.insert(freemem_t::value_type(object->header.size, PageOffsetOfAddr(addr)));
}
