/* ptr.h -*- C++ -*- automatic ptr class
 * Rob Earhart
 * $Id: ptr.h,v 1.1.1.1 1997/11/26 21:00:57 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_PTR_H
#define ACAP_PTR_H

// It's difficult to maintain the "initialization == aquisition" model,
// much less provide automatic destruction of non-garbage collected objects,
// without some form of automatic pointer template, whose sole purpose
// in life is to delete what it points to when it's destroyed if it
// in fact points to something.

// Note that when we assign from another Ptr, the other Ptr points to NULL;
// this is a little nasty, but ensures that, when Ptrs are used
// exclusively, only one of them will point to a given object, and thus
// the object will only be deleted once; this is considered desireable. :-)

// Get bool
#include <stl_config.h>

#ifndef NULL
#define NULL (0)
#endif

template<class T>
class Ptr {
public:
  Ptr() : t(NULL) {}
  Ptr(T *_t) : t(_t) {}
  Ptr(const Ptr<T> &it) : t(it.t) {
    // I'd use const_cast<> here, but Solaris's C++ compiler
    // doesn't seem to grok it yet.  :-(
    ((Ptr<T> &)(it)).t = NULL;
  }
  ~Ptr() {
    if (t) delete t;
  }
  T *Value() const { return t; }
  operator T *() const { return t; }
  T &operator*() const { return *t; }
  T *operator->() const { return t; }
  bool operator!() const { return ! t; }
  Ptr<T> &operator=(T *_t) {
    if (t && t != _t) {
      delete t;
    }
    t = _t;
    return *this;
  }
  Ptr<T> &operator=(const Ptr<T> &it) {
    if (t && t != it.t) {
      delete t;
    }
    t = it.t;
    // Another place where const_cast<> would be nice...
    ((Ptr<T>&)(it)).t = NULL;
    return *this;
  }
  void Nullify() { t = NULL; }	// special case - doesn't delete old t
private:
  T *t;
};

// APtr's used for arrays; objects that should be deleted via "delete []".
// It lacks the -> operator (which isn't well-defined for arrays),
// and has a [] operator.

template<class T>
class APtr {
public:
  APtr() : t(NULL) {}
  APtr(T *_t) : t(_t) {}
  APtr(const APtr<T> &it) : t(it.t) {
    // Another place where const_cast<> would be nice...
    ((APtr<T>&)(it)).t = NULL;
  }
  ~APtr() {
    if (t) delete [] t;
  }
  T *Value() const { return t; }
  operator T *() const { return t; }
  bool operator!() const { return ! t; }
  APtr<T> &operator=(T *_t) {
    if (t && t != _t) {
      delete [] t;
    }
    t = _t;
    return *this;
  }
  APtr<T> &operator=(const APtr<T> &it) {
    if (t && t != it.t) {
      delete [] t;
    }
    t = it.t;
    // Another place where const_cast<> would be nice...
    ((APtr<T>&)(it)).t = NULL;
    return *this;
  }
  void Nullify() { t = NULL; }	// special case - doesn't delete old t
private:
  T *t;
};

#endif // ACAP_PTR_H
